"""Data loading for the [OSSL dataset](https://explorer.soilspectroscopy.org/)"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_datasets.ossl.ipynb.

# %% auto 0
__all__ = ['OSSLData', 'get_cache_path', 'get_ossl']

# %% ../../nbs/01_datasets.ossl.ipynb 2
from fastcore.all import *
from pathlib import Path
import pandas as pd
from urllib.request import urlretrieve
import re
import numpy as np

from ..types import *

# %% ../../nbs/01_datasets.ossl.ipynb 3
class OSSLData:
    def __init__(self, df):
        self.df = df
        self._parse_columns()        
        self.sample_ids = (df['id.layer_local_c'].values if 'id.layer_local_c' 
                           in df.columns else np.arange(len(df)))

# %% ../../nbs/01_datasets.ossl.ipynb 4
@patch
def _parse_columns(self:OSSLData):
    "Parse columns into visnir, mir and properties"
    self.visnir_cols = [c for c in self.df.columns if re.match(r'scan_visnir\.\d+_ref', c)]
    self.mir_cols = [c for c in self.df.columns if re.match(r'scan_mir\.\d+_abs', c)]
    spectral_cols = set(self.visnir_cols + self.mir_cols)
    self.properties_cols = [c for c in self.df.columns if c not in spectral_cols]

# %% ../../nbs/01_datasets.ossl.ipynb 5
def get_cache_path(): return Path.home()/'.soilspecdata'

# %% ../../nbs/01_datasets.ossl.ipynb 6
def get_ossl(url='https://storage.googleapis.com/soilspec4gg-public/ossl_all_L0_v1.2.csv.gz', # OSSL URL
             force_download=False # if True, force download
             ):
    "Get OSSL data"
    cache_path = get_cache_path()/'ossl_v1.2.csv.gz'
    if not cache_path.exists() or force_download:
        cache_path.parent.mkdir(exist_ok=True)
        urlretrieve(url, cache_path)
        
     # Define date columns
    date_columns = [
        'scan.mir.date.begin_iso.8601_yyyy.mm.dd',
        'scan.mir.date.end_iso.8601_yyyy.mm.dd',
        'scan.visnir.date.begin_iso.8601_yyyy.mm.dd',
        'scan.visnir.date.end_iso.8601_yyyy.mm.dd'
    ]
    
    # Update dtype dictionary without datetime columns
    dtype = {
        # IDs and codes
        'id.layer_local_c': 'string',
        'id.location_olc_txt': 'string',
        'id.dataset.site_ascii_txt': 'string',
        'id.scan_local_c': 'string',
        
        # Categorical text fields
        'layer.texture_usda_txt': 'category',
        'pedon.taxa_usda_txt': 'category',
        'horizon.designation_usda_txt': 'category',
        'location.country_iso.3166_txt': 'category',
        'surveyor.address_utf8_txt': 'category',
        'efferv_usda.a479_class': 'category',
        
        # Text fields
        'scan.mir.model.name_utf8_txt': 'string',
        'scan.mir.model.code_any_txt': 'string',
        'scan.mir.method.optics_any_txt': 'string',
        'scan.mir.method.preparation_any_txt': 'string',
        'scan.mir.license.title_ascii_txt': 'string',
        'scan.mir.license.address_idn_url': 'string',
        'scan.mir.doi_idf_url': 'string',
        'scan.mir.contact.name_utf8_txt': 'string',
        'scan.mir.contact.email_ietf_txt': 'string',
        'scan.visnir.model.name_utf8_txt': 'string',
        'scan.visnir.model.code_any_txt': 'string',
        'scan.visnir.method.optics_any_txt': 'string',
        'scan.visnir.method.preparation_any_txt': 'string',
        'scan.visnir.license.title_ascii_txt': 'string',
        'scan.visnir.license.address_idn_url': 'string',
        'scan.visnir.doi_idf_url': 'string',
        'scan.visnir.contact.name_utf8_txt': 'string',
        'scan.visnir.contact.email_ietf_txt': 'string'
    }
    df = pd.read_csv(cache_path, compression='gzip', dtype=dtype,
                     parse_dates=date_columns)
    return OSSLData(df)

# %% ../../nbs/01_datasets.ossl.ipynb 9
@patch
def _get_valid_spectra_mask(self:OSSLData, spectra_cols):
    """Return mask for samples with all non-null values in spectra"""
    return self.df[spectra_cols].notna().all(axis=1)

# %% ../../nbs/01_datasets.ossl.ipynb 11
@patch
def _extract_wavenumbers(self:OSSLData, 
                        cols: List[str] # column names
                        ):
    "Extract wavenumbers from column names"
    return np.array([int(re.search(r'\.(\d+)_', c).group(1)) for c in cols])

# %% ../../nbs/01_datasets.ossl.ipynb 13
@patch
def _extract_measurement_type(self:OSSLData, 
                              cols: List[str] # column names
                              ):
    "Extract measurement type from column names"
    types = set(re.search(r'_(\w+)$', c).group(1) for c in cols)
    assert len(types) == 1, f"Mixed measurement types found: {types}"
    return types.pop()

# %% ../../nbs/01_datasets.ossl.ipynb 15
@patch
def _filter_wavelength_range(self:OSSLData, 
                             wavenumbers: np.ndarray, # wavenumbers
                             spectra: np.ndarray, # spectra
                             cols: List[str], # column names
                             wmin: Optional[float]=None, # min wavenumber
                             wmax: Optional[float]=None # max wavenumber
                             ):
    "Filter spectra based on wavelength range"
    mask = np.ones(len(wavenumbers), dtype=bool)
    if wmin is not None:
        mask &= wavenumbers >= wmin
    if wmax is not None:
        mask &= wavenumbers <= wmax
    return wavenumbers[mask], spectra[:, mask], [cols[i] for i in np.where(mask)[0]]

# %% ../../nbs/01_datasets.ossl.ipynb 18
@patch 
def get_visnir(self:OSSLData, 
               wmin: Optional[float]=None, # min wavenumber
               wmax: Optional[float]=None, # max wavenumber
               require_valid: bool=True # if True, only return samples with no null values
               ):
    "Get VISNIR spectra within specified wavenumber range"
    wavenumbers = self._extract_wavenumbers(self.visnir_cols)
    spectra = self.df[self.visnir_cols].values
    wavenumbers, _, filtered_cols = self._filter_wavelength_range(
        wavenumbers, spectra, self.visnir_cols, wmin, wmax
    )
    
    if require_valid:
        valid_mask = self._get_valid_spectra_mask(filtered_cols)
        df_subset = self.df[valid_mask]
        sample_ids = self.sample_ids[valid_mask]
    else:
        df_subset = self.df
        sample_ids = self.sample_ids
        
    spectra = df_subset[filtered_cols].values
    measurement_type = self._extract_measurement_type(filtered_cols)
    return SpectraData(wavenumbers, spectra, measurement_type, sample_ids)

# %% ../../nbs/01_datasets.ossl.ipynb 20
@patch 
def get_mir(self:OSSLData, 
            wmin: Optional[float]=600, # min wavenumber
            wmax: Optional[float]=4000, # max wavenumber
            require_valid: bool=True # if True, only return samples with no null values
            ):
    "Get MIR spectra within specified wavenumber range"
    wavenumbers = self._extract_wavenumbers(self.mir_cols)
    spectra = self.df[self.mir_cols].values
    wavenumbers, _, filtered_cols = self._filter_wavelength_range(
        wavenumbers, spectra, self.mir_cols, wmin, wmax
    )
    
    if require_valid:
        valid_mask = self._get_valid_spectra_mask(filtered_cols)
        df_subset = self.df[valid_mask]
        sample_ids = self.sample_ids[valid_mask]
    else:
        df_subset = self.df
        sample_ids = self.sample_ids
        
    spectra = df_subset[filtered_cols].values
    measurement_type = self._extract_measurement_type(filtered_cols)
    
    return SpectraData(wavenumbers, spectra, measurement_type, sample_ids)

# %% ../../nbs/01_datasets.ossl.ipynb 22
@patch
def get_properties(self:OSSLData, 
                   properties=None, # properties
                   require_complete: bool=False # if True, only return samples with no null values
                   ):
    "Get properties data with sample IDs"
    if properties is None:
        properties = self.properties_cols
    elif isinstance(properties, str):
        properties = [properties]
            
    df_subset = pd.DataFrame({
        'id': self.sample_ids,
        **{col: self.df[col] for col in properties}
    }).set_index('id')
        
    if require_complete:
        return df_subset.dropna()
    return df_subset

# %% ../../nbs/01_datasets.ossl.ipynb 29
@patch
def get_aligned_data(self:OSSLData, 
                    spectra_data: SpectraData, # spectra data
                    target_cols: Union[str, List[str]] # target columns
                    ): 
    "Get aligned spectra and target data for ML, along with their sample IDs."
    # Get targets with complete data
    targets = self.get_properties(target_cols, require_complete=True)
    
    # Find common IDs between spectra and targets
    common_ids = list(set(spectra_data.sample_ids) & set(targets.index))
    
    # Create index mapping for efficient lookup
    spectra_id_to_idx = {id_: idx for idx, id_ in enumerate(spectra_data.sample_ids)}
    
    # Get indices for alignment
    indices = [spectra_id_to_idx[id_] for id_ in common_ids]
    
    # Align the data
    features = spectra_data.spectra[indices]
    targets = targets.loc[common_ids].values
    sample_ids = np.array(common_ids)
    
    return features, targets, sample_ids
